/****************************************************************************************[Dimacs.h]
Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
Copyright (c) 2007-2010, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#ifndef Glucose_Dimacs_h
#define Glucose_Dimacs_h

#include <map>
#include <stdio.h>

#include "utils/ParseUtils.h"
#include "core/SolverTypes.h"

namespace Glucose {

//=================================================================================================
// DIMACS Parser:

template<class B, class Solver>
static void readClause(B& in, Solver& S, vec<Lit>& lits, int var1, int vars) {
    int     parsed_lit, var;
    lits.clear();
    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        //while (var >= S.nVars()) S.newVar();
        if (var >= vars)
            printf("c ERROR! DIMACS header mismatch: wrong number of variables.\n"), exit(3);
        lits.push( (parsed_lit > 0) ? mkLit(var1 + var) : ~mkLit(var1 + var) );
    }
}

template<class B, class Solver>
static void parse_DIMACS_main(B& in, Solver& S, int& var1, int& vars, bool& minimize, std::map<int, int>& obj_term) {
    vec<Lit> lits;
    var1 = S.nVars();
    vars = 0;
    int clauses = 0;
    int cnt     = 0;
    bool wcnf   = false;
    int weight  = 1;
    int ub      = static_cast<int>(minimize) + 1;
    //for (;;){
        skipWhitespace(in);
        if (*in == EOF)
            printf("c PARSE ERROR! Unexpected EOF\n"), exit(3); //break;
        else if (*in == 'p'){
            ++in;
            if (*in != ' ')
                goto UnexpectedChar;
            ++in;
            if (*in == 'c'){
                if (eagerMatch(in, "cnf")){
                    vars    = parseInt(in);
                    clauses = parseInt(in);
                    // SATRACE'06 hack
                    // if (clauses > 4000000)
                    //     S.eliminate(true);
                }else
                    goto UnexpectedChar;
            }
            else if (*in == 'w'){
                if (eagerMatch(in, "wcnf")){
                    vars    = parseInt(in);
                    clauses = parseInt(in);
                    wcnf    = true;
                    ub      = parseInt(in);
                    minimize = true;
                    // SATRACE'06 hack
                    // if (clauses > 4000000)
                    //     S.eliminate(true);
                }else
                    goto UnexpectedChar;
            }else
                goto UnexpectedChar;
            for (int i = 0; i < vars; ++i)
                S.newVar();
        }else{
        UnexpectedChar:
            printf("c PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
        }
    // loop now here so we don't accidentally interpret header 0 or 2+ times
    for (;;){
        skipWhitespace(in);
        if (*in == EOF) break;
        else if (*in == 'c' || *in == 'p')
            skipLine(in);
        else{
            cnt++;
            if (wcnf)
                weight = static_cast<unsigned>(parseInt(in));
            readClause(in, S, lits, var1, vars);
            if (weight < ub){
                Lit p = mkLit(S.newVar(), false);
                lits.push(p);
                obj_term.insert(std::pair<int, int>(toInt(p), weight));
            }
            S.addClause_(lits); }
    }
    //if (vars != S.nVars())
    //    printf("c WARNING! DIMACS header mismatch: wrong number of variables.\n");
    if (cnt  != clauses)
        printf("c WARNING! DIMACS header mismatch: wrong number of clauses.\n");
}

// Inserts problem into solver.
//
template<class Solver>
static void parse_DIMACS(gzFile input_stream, Solver& S, int& var1, int& vars, bool& minimize, std::map<int, int>& obj_term) {
    StreamBuffer in(input_stream);
    parse_DIMACS_main(in, S, var1, vars, minimize, obj_term); }

//=================================================================================================
}

#endif
