/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
 *  Main authors:
 *     Ignasi Abio <iabio@nicta.com.au>
 */

/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include <core/Solver.h>
#include <toSAT/AMO/Heule.hh>

namespace Glucose {

  inline void Heule_quadratic(Solver* s, const std::vector<Lit>& V, Lit halfreif/*Var e*/) {
    for (uint i = 0; i < V.size(); i++) {
      for (uint j = i+1; j < V.size(); j++) {
	vec<Lit> v(3);
	v[0]=V[i];
	v[1]=V[j];
	v[2]=~halfreif; //mkLit(e, true);
	s->addClause(v);
      }
    }
  }

  void/*Var*/ Heule_AMO(Solver* s, const std::vector<Lit>& x, intptr_t parameters, Lit halfreif) {
    uint lambda = (uint)parameters;
    //Var ret = s->newVar();
    std::vector<Lit> x2(x);

    while(x2.size() > lambda) {
      std::vector<Lit> x3(lambda);
      for (uint i = 0; i < lambda-1; i++) {
	x3[i]=x2.back();
	x2.pop_back();
      }
      Var e=s->newVar();
      x2.push_back(mkLit(e, false));
      x3.back()=mkLit(e, true);
      Heule_quadratic(s, x3, halfreif/*ret*/);
    }
    Heule_quadratic(s, x2, halfreif/*ret*/);
    //return ret;
  }

};
