/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
 *  Main authors:
 *     Ignasi Abio <iabio@nicta.com.au>
 */

/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include <vector>
#include <core/Solver.h>
#include <toSAT/LI/MDD_LI.hh>

namespace Glucose {

  void/*Var*/ MDD_leqLI(Solver* s, const std::vector<int>& c, const std::vector<std::vector<Lit> >& x,
			int b, intptr_t parameters, Lit halfreif) {

    // std::cout << "LI:";
    // for (unsigned int i=0; i<c.size(); i++) std::cout << "+" << c[i] << " *" << x[i].size() << " ";
    // std::cout << "<= " << b << std::endl;

    double lambda = (uint)parameters;
    //std::vector<Lit> toAdd;
    std::vector<std::vector<Lit> > x2(x);
    std::vector<int> c2(c);
    //std::cout << "sorting" << std::endl;
    LI_Commons::sort(x2, c2, false);
    //std::cout << "sorted" << std::endl;

    std::vector<std::vector<Lit> > x3;
    std::vector<int> c3;
    for (uint i=0; i<x2.size(); i++) {
      uint ini = i;
      while( i < x2.size()-1 && c2[i] == c2[i+1] ) i++;
      if (c2[i] == 0 ) continue;
      std::vector<Lit> V(x2[ini]);
      for (uint j = ini+1; j <= i; j++) {

        //std::cout << "Merging " << V.size() << " and " << x2[j].size() << std::endl;

	std::vector<Lit> aux;
	std::vector<Lit> aux2(x2[j]);
	GetSimpMerge::Instance(lambda).createNetwork(s,V,aux2,aux,b/(c2[ini])+1);
	V = aux;
      }
      while ( V.size() * c2[ini] > b ) {
        vec<Lit> v(2);
        v[0] = ~halfreif;
        v[1] = V.back();
        s->addClause(v);
	//toAdd.push_back(V.back());
	V.pop_back();
      }
      x3.push_back(V);
      c3.push_back(c2[ini]);
    }
    int t=0;
    for (uint i = 0; i < x3.size(); i++) t+=c3[i]*x3[i].size();
    if ( t <= b )
      return; //{/*std::cout << "ret" << std::endl;*/ return ret(s,toAdd); }

    // std::cout << "--- LI:";
    // for (unsigned int i=0; i<c3.size(); i++) std::cout << "+" << c3[i] << " *" << x3[i].size() << " ";
    // std::cout << "<= " << b << std::endl;
 
    std::vector<int> dom(x3.size());
    for(uint i = 0; i < x3.size(); i++) dom[i]=x3[i].size();
    MDD mdd;
    mdd.defineConstraint(c3, dom);

    NodeID node;
    NodeID root = mdd.newRoot(b);
    if ( mdd.isTerminal(root) ) {
      //Var r = ret(s,toAdd);
      if (!mdd.isTrue(root)) {
	vec<Lit> v(1); //, mkLit(r, true));
        v[0] = ~halfreif;
	s->addClause(v);
      }
      return; //r;
    }
    node = mdd.getFirstNode();
    while(true) {
      if (!mdd.isTerminal(node)) {
	uint nS = mdd.getNumSons(node);
	if ( mdd.getSon(node,0) == mdd.getSon(node, nS-1) ) { // Long edge node.
	  if ( mdd.isTerminal(mdd.getSon(node,0)) )	mdd.setTerminal(node, mdd.isTrue(mdd.getSon(node,0)) );
	  else mdd.setVar(node,mdd.getVar(mdd.getSon(node,0)));
	} else {
	  Var nodeVar=node == root ? var_Undef : s->newVar();
	  NodeID last = node; // No last.
	  for (uint i = 0; i < nS; i++) {
	    NodeID current = mdd.getSon(node,i);
	    if ( last != current ) {
	      last = current;
	      if (mdd.isTerminal(current)) {
		if (!mdd.isTrue(current)) {
		  // Current is the false node. Clause "node -> lit < i" is needed (or, equiv., 'NEG n' OR 'lit <= i-1').
		  vec<Lit> v(2);
 assert(i > 0);
                  v[0]=~halfreif;
                  v[1]=x3[mdd.getLevel(node)][i-1];
                  if (nodeVar != var_Undef)
		    v.push(mkLit(nodeVar, true));
		  s->addClause(v);
		} // No else: if current is the true node, nothing is needed.
	      } else {
		// General case: "n AND 'lit >= i' -> current"
		vec<Lit> v(1);
		v[0]=mkLit(mdd.getVar(current), false);
                if (nodeVar != var_Undef)
		  v.push(mkLit(nodeVar, true));
		if (i > 0)
		  v.push(x3[mdd.getLevel(node)][i-1]); // Note: if i = 0, 'lit >= i' is a tautology
		s->addClause(v);
	      }
	    }
	  }
	  mdd.setVar(node, nodeVar);
	}
      }
      if (mdd.isLast(node)) break;
      else node = mdd.nextNode(node);
    }
    return; //ret(s,toAdd,mdd.getVar(root));

  }


  //Var ret(Solver* s, const std::vector<Lit>& toAdd, Var r) {
  //  if ( r == var_Undef )
  //    r = s->newVar();
  //  vec<Lit> v(2);
  //  v[0]=mkLit(r, true);
  //  for (uint i=0; i<toAdd.size(); i++) {
  //    v[1]=toAdd[i];
  //    s->addClause(v);
  //  }
  //  return r;
  //}

};
