/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
 *  Main authors:
 *     Ignasi Abio <iabio@nicta.com.au>
 */

/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef __MINIZINC_INTERVAL_HH__
#define __MINIZINC_INTERVAL_HH__

#include <vector>

typedef uint NodeID;

class Interval {
private:
  int a, b;
  NodeID n;
public:
  inline Interval ( const std::vector<Interval> &vi, int add, NodeID _n );
  inline Interval ( int _a );
  inline Interval ( bool b, NodeID _n );
  inline Interval ( const Interval& i );
  inline bool operator < ( const Interval &i2 ) const;
  inline bool operator > ( const Interval &i2 ) const;
  inline bool operator == ( const Interval &i2 ) const;
  inline void print () const;
  inline int getA () const { return a; }
  inline int getB () const { return b; }
  inline NodeID getNodeID () const { return n; }
};

////////////////////////////////////////////////////////////////////////////
////////////////////// Interval Class Implementation: //////////////////////
////////////////////////////////////////////////////////////////////////////

inline Interval::Interval ( const std::vector<Interval> &vi, int add, NodeID _n ):n(_n) {
  //assert( vi.size()>0 && vi[0].a <= vi[0].b );
  a=vi[0].a; b=vi[0].b;
  int x = add;
  for (uint i = 1; i < vi.size(); i++){
    int viax = (vi[i].a == INT_MIN? INT_MIN:vi[i].a+x);
    int vibx = (vi[i].b == INT_MAX? INT_MAX:vi[i].b+x);

    //assert( vi[i].a <= vi[i].b && viax <= b && vibx >= a );
    if (viax > a) a = viax;
    if (vibx < b) b = vibx;
    x+=add;
  }
}
inline Interval::Interval ( int _a ):a(_a),b(_a) {}
inline Interval::Interval ( bool _b, NodeID _n ):n(_n) {
  a=( _b ? 0      : INT_MIN );
  b=( _b ? INT_MAX : -1     );
}
inline Interval::Interval ( const Interval& i ):a(i.a),b(i.b),n(i.n) { /*assert(a <= b);*/ }
inline bool Interval::operator < (const Interval &i2 ) const { return b < i2.a; }
inline bool Interval::operator > (const Interval &i2 ) const { return a > i2.b; }
inline bool Interval::operator == (const Interval &i2 ) const { return ( b >= i2.a && a <= i2.b); }
/*inline void Interval::print() const {
  if (a == INT_MIN) cout << "( -inf, "; else cout << "[ " << a << ", ";
  if (b == INT_MAX) cout << "inf )" << endl; else cout << b << " ]" << endl;
}*/

#endif
